// ===============================================================================
//						  AGEIA PHYSX SDK TRAINING PROGRAMS
//					            PARTICLE SYSTEM CLASS
//
//						  Written by Sam Thompson, 3-17-05
//						    Edited by Bob Schade, 5-1-06
// ===============================================================================

#include "Particle.h"

Particle::Particle()
{
	setActor(NULL);
	_vThermoDynamicForce.zero();
}

Particle::Particle(NxActor* pToSet, NxVec3& vThermoDynamicAcceleration)
{
	setActor(pToSet);

	// F = M*a
	_vThermoDynamicForce = pToSet->getMass()*vThermoDynamicAcceleration;
}

void Particle::setActor(NxActor* pToSet)
{
	_pActor = pToSet;
}

void Particle::update(NxReal fDeltaTime, NxReal fMaxRadius, NxReal fGrowthRate)
{
	// particles have a false, thermodynamic force applied which pushes them up every frame
	_pActor->addForce(_vThermoDynamicForce,NX_FORCE,false);

	if (_pActor->isSleeping()==false)
	{
		NxVec3 ppos = _pActor->getGlobalPosition();
		NxQuat q = _pActor->getGlobalOrientationQuat();

	    pEntity[13] = q.x;
		pEntity[14] = q.y;
		pEntity[15] = q.z;
		pEntity[12] = -q.w;
		pEntity[16] = ppos.x;
		pEntity[17] = ppos.y;
		pEntity[18] = ppos.z;
		int moved = 3;
		pEntity[11] = *((float*)(&moved));
	}

	// Use our down-caster to get a sphere pointer (and make sure it is a sphere)
	NxSphereShape* pSphere = _pActor->getShapes()[0]->isSphere();
	//const NxReal fMaxRadius = 3.0f;
	//const NxReal fGrowthRate = 0.6f;
	if (pSphere && fGrowthRate)
	{
		// Grow the radius at growth rate fGrowthRate m/s, until it reaches fMaxRadius meters
		NxReal fCurrentRadius = pSphere->getRadius();
		if (fCurrentRadius < fMaxRadius)
		{
			pSphere->setRadius(pSphere->getRadius() + fGrowthRate/fDeltaTime);
			if (pSphere->getRadius() > fMaxRadius) pSphere->setRadius(fMaxRadius);
		}
		else if (fCurrentRadius > fMaxRadius)
		{
			pSphere->setRadius(pSphere->getRadius() - fGrowthRate/fDeltaTime);
			if (pSphere->getRadius() < fMaxRadius) pSphere->setRadius(fMaxRadius);
		}
	}
}

NxActor& Particle::getActorReference( )
{
	return *_pActor;
}

ParticleEmitter::ParticleEmitter(NxScene* pSceneToSet, NxVec3 vStartingPos, NxReal fStartingVelScale, NxVec3 vThermoDynamicAcceleration)
{
	_iHeadIndex = 0;
	_iTailIndex = 0;
	_iParticleCount = 0;
	_pScene = pSceneToSet;
	_vStartingPos = vStartingPos;
	_fNewParticleTimer = 0.0f;
	_vThermoDynamicAcceleration = vThermoDynamicAcceleration;
	_peOrientation.zero();
	_pMass = 1.0f;
	_pADamping = 5.0f;
	_pLDamping = 0.2f;
	_pRadius = 1.0f;
	_pScaleRadius = 1.0f;
	_pScaleStartVelmin = 1.0f;
	_pScaleStartVelmax = 10.0f;
	_fGrowthRate = 0.0f;
}

ParticleEmitter::~ParticleEmitter()
{
    removeAllParticles();
}

void ParticleEmitter::removeAllParticles()
{
	while (_iParticleCount > 0)
		removeParticle();
}

void ParticleEmitter::update(NxReal fDeltaTime)
{
	_fNewParticleTimer -= fDeltaTime;
	if (_fNewParticleTimer < 0.0f)
	{
	//	addParticle();
	//	_fNewParticleTimer += cfTimeBetweenParticles;
	}

	// Update all particles
	int iParticlesUpdated = 0;
	for (int iParticleIndex = _iTailIndex; iParticlesUpdated < _iParticleCount; iParticleIndex = ((iParticleIndex+1) % ciMaxParticles))
	{
		_aParticleArray[iParticleIndex]->update(fDeltaTime, _fMaxRadius, _fGrowthRate);
		++iParticlesUpdated;
	}
}

Particle* ParticleEmitter::addParticle()
{

	if ((_iTailIndex == _iHeadIndex) && (_iParticleCount != 0)) // FIFO is full
	{
		removeParticle();
		// Now there is a slot free
	}
	// Add a single-shape actor to the scene
	NxActorDesc actorDesc;
	NxBodyDesc bodyDesc;
	bodyDesc.mass = _pMass;
	bodyDesc.angularDamping = _pADamping;
	bodyDesc.linearDamping = _pLDamping;
	bodyDesc.sleepLinearVelocity = 1.0f;
	//bodyDesc.linearDamping = 0.5f;

	NxSphereShapeDesc sphereDesc;
	sphereDesc.radius = _pRadius;
	sphereDesc.group = cParticleCollisionGroup; // this group does not collide with itself
	actorDesc.shapes.pushBack(&sphereDesc);

	actorDesc.body = &bodyDesc;
	actorDesc.globalPose.t = _vStartingPos;

	NxActor* pNewActor =  _pScene->createActor(actorDesc);
	
	NxGroupsMask GrMask;
//	mainSetMask(pNewActor);
//	BodyAddUserData(pNewActor);

	// Give it an initial linear velocity, scaled by _fStartingVelScale
	NxVec3 vRandVec(NxReal(NxMath::rand(-_pScaleRadius,_pScaleRadius)), 
		NxReal(NxMath::rand(_pScaleStartVelmin,_pScaleStartVelmax)), 
		NxReal(NxMath::rand(-_pScaleRadius,_pScaleRadius)));

	_peOrientation.rotate(vRandVec);
	pNewActor->setLinearVelocity(vRandVec);

	// Turn off gravity for smoke
	pNewActor->raiseBodyFlag(NX_BF_DISABLE_GRAVITY);

	Particle* mParticle = new Particle(pNewActor, _vThermoDynamicAcceleration);
	_aParticleArray[_iHeadIndex] = mParticle;
	mParticle->_pIndex = _iHeadIndex;

	_iHeadIndex = (_iHeadIndex+1) % ciMaxParticles;
	++_iParticleCount;

	return mParticle;
}

float* ParticleEmitter::removeParticle( )
{
	// release the actor the tail particle owns
	float* entity = _aParticleArray[_iTailIndex]->pEntity;
	_pScene->releaseActor(_aParticleArray[_iTailIndex]->getActorReference());
	// delete it
	delete _aParticleArray[_iTailIndex];
	// increment the tail
	_iTailIndex = (_iTailIndex+1) % ciMaxParticles;
	--_iParticleCount;
	return entity;
}
void ParticleEmitter::removeOneParticle(Particle* particle)
{
	// release the actor the tail particle owns
	int Index = particle->_pIndex;
	_pScene->releaseActor(_aParticleArray[Index]->getActorReference());
	// delete it
	delete _aParticleArray[Index];
	_iTailIndex = (_iTailIndex+1) % ciMaxParticles;
	// increment the tail
	--_iParticleCount;
}